/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#if !defined(_SOL_DICT)
#define _SOL_DICT
#define REVERSE true

class sol_pair {
 public:
 text key;
 dyn  *value;

 sol_pair() { 
     key = text(); 
     value = new dyn(); 
 }

 sol_pair(sol_pair *s) {
   key   = s->key;
   value = s->value;
 }

 sol_pair(int a, dyn b)  {
  key = text(a);
  value = new dyn(b);
 }

 sol_pair(const char *a, dyn b)  {
  key = text(a);
  value = new dyn(b);
 }

 sol_pair(text a, dyn b)  {
  key = a;
  value = new dyn(b);
 }

 text toText();
 const char *toStr();

};


/**
  Associative, Php compatible dynamic solsolarray.
  Base class also of dict.
  */

class dict {
 std::vector<text> keyList;
 std::vector<dyn> data;
 static dyn nullelement;	
	
 public:

 int indexer;

 /**
   Return the number of elements
   */
 inline int size()   { return (int) data.size(); }

 /**
   Return true if the solsolarray holds no any element, false otherwise
   */
 inline bool empty() { return data.empty(); }

 void clear();

 void set(const dict &);
 void set(dict *);
 void set(dyn);
 void set(std::map<text, dyn> &);
 inline void set(std::vector<dyn> *a) { set(&a); }

 void setAt(const text &, dyn);
 void setAt(const text &, dyn *);


 dict();

 /**
   Build an associative dict from an dict of strings of chars.
   Indexes are generated from 0 to n.
   */
 dict(char *[]);

 /**
   Build an dict with a single element.
   */
 dict(dyn);
 
 /**
  Clone
  */
 
 dict(const dict &);
 dict(dict *);
 
 /** From a std map
  */

 dict(std::map<text, dyn> &);
 dict(std::map<text, dyn> *);
 

 ~dict();

 dyn &operator[](const char *);
 dyn &operator[](const text &);

 void operator=(const dict &);
 void operator=(dict *);
 void operator=(dyn);
 void operator=(std::map<text, dyn> &);


 void add(dict *);
 inline void add(const dict &d) { add(d); }

 dyn get(const text &);
 dyn get(const char *);
 dyn get(int);

 inline dict operator+(dict x) { return operator+(&x); }
 dict operator+(dict *);
 dict operator+(dyn);

 inline dict operator-(dict x) { return operator-(&x); }
 dict operator-(dict *);
 dict operator-(dyn);
 
 inline dict operator&(dict x) { return operator&(&x); }
 dict operator&(dict *);
 
 inline dict operator|(dict x) { return operator|(&x); }
 dict operator|(dict *);
 
 inline dict operator^(dict x) { return operator^(&x); }
 dict operator^(dict *);

 inline dict &operator+=(dict x) { return operator+=(&x); }
 dict &operator+=(dict *);
 dict &operator+=(dyn);
 
 void operator-=(dict x) { operator-=(&x); }
 void operator-=(dict *);
 void operator-=(dyn);
 
 inline bool operator==(dict x) { return operator==(&x); }
 bool operator==(dict *);
 bool operator==(bool);
 
 inline bool operator!=(dict x) { return operator!=(&x); }
 bool operator!=(dict *);
 bool operator!=(bool);
 
 inline bool operator<(dict x) { return operator<(&x); }
 bool operator<(dict *);
 
 inline bool operator>(dict x) { return operator>(&x); }
 bool operator>(dict *);
 
 inline bool operator>=(dict a) { return !operator<(&a); }
 inline bool operator>=(dict *a) { return !operator<(a); }
 
 inline bool operator<=(dict b)  { return !operator>(&b); }
 inline bool operator<=(dict *b) { return !operator>(b); }
 
 inline void operator&=(dict x) { operator&=(&x); }
 void operator&=(dict *);
 
 inline void operator|=(dict x) { operator|=(&x); }
 void operator|=(dict *);
 
 inline void operator^=(dict x) { operator^=(&x); }
 void operator^=(dict *);
 
 bool operator!();
 dict &operator,(dyn);

 //===== util
 bool in(const dyn &);
 bool in(dyn *);
 signed int hasStrKey(const text &);
 signed int findKey(const text &);
 signed int findKey(const char *);
 signed int findKey(int);
 signed int findValue(const dyn &);
 signed int findValue(dyn *);
 inline int find(dyn x)  { return findValue(x); }
 inline int find(dyn *x) { return findValue(x); }
 signed int findLastValue(const dyn &);
 signed int findLastValue(dyn *);
 inline int findLast(dyn x)  { return findLastValue(x); }
 inline int findLast(dyn *x) { return findLastValue(x); }
 void display();
 void display(int);
 void pushKey(const text &);

 //===== iterator

 /**
   Restart the iterator (at first element).
   */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();
 /**
   Return the current element, and move index to next one.
   */
 dyn inc();

 /**
   Return the current element, and move index to previous one.
   */
 dyn dec();

 /**
   Return the currently indiced element.
   */
 dyn current();

 /**
   Return the key of the currently pointed out element.
   */
 text key();

 /**
   Return the key of the element at position in argument.
   */
 text key(int);

 /**
   Return the value of the currently pointed out element.
   */
 dyn value();


 /**
   Return the value of the element at position in argument.
   */
 dyn value(int);

 /**
   Set iterator at first element and return it.
   */
 dyn begin();

 /**
   Set iterator at last element and return it.
   */
 dyn end();

 /**
   Return the current indice of the iterator.
   */
 int index();

 //===== stack

 /**
   Push a dyn on top of the dict, it becomes the last element
   and the size increases from one.
   */
 void push(const dyn &);
 void push(dyn *);
 void push(sol_pair *);
 void push(const text &, dyn);

 void pushAdd(const text &, dyn);
 void pushAdd(int, dyn);
 void pushUnic(const text &, dyn);

 /**
   Return the last element, and remove it from the dict.
   If empty, return false.
   */
 dyn  pop();

 /**
   Return the first element and remove it.
   All remaining element are shift for one position. Indices are
   renumbered if integers.
   */
 dyn  shift();

 /**
   Insert an element at the first position.
   */
 void unshift(text, dyn);

 /**
   Insert an element at given position.
   The first parameter is the postion, the second one the element.
   */

 void insert(int, dyn);
 
 //===== ranges
 void erase(int);
 void erase(int, int);
 void erase(const char *);
 inline void erase(const text t) { erase(t.data); }
 

 dict slice(int, int, int);
 
 void splice(int, int, dict, int);
 void splice(int, int, dict *, int);
 void splice(int, int, int);
 void splice(int, int);
 void splice(int);
 void splice(int, int, dyn, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Renumber the key if they are numbers. Textual keys are ignored.
   */
 void pack();

 void swapData(int, int);
 void swapElem(int, int);
 void keysSort(int, int);
 void dynSort(int, int);
 void dynRSort(int, int); 

 /**
   Reverse the content of the dict
   */
 dict reverse();

 /**
   Sort the values of the dict in ascending order.
   */
 void sort(int x = 0);
 void rSort(int x = 0);

 /**
   Sort the keys of the dict, preserving the associations with values.
   Ascending order.
   */
 void kSort(int x = 0);

 /**
   Remove doubloons in values.
   */
 dict unique();

 /**
   Return the values as an dict. Keys are generated, from 0 to n.
   */
 solarray values();

 /**
   Return the keys as an dict. Integer keys are generated.
   */
 solarray keys();

 //===== operations
 void clone(dict *);   // copy to another dict

  /**
   Return the minimal value among elements.
   */
 dyn min();

 /**
   Return items selected by the function.
   */
 dict map(bool (*)(dyn));

 /**
   Return the greatest values among elements.
   */
 dyn max();

 /**
   Calculate the sum of elements and return it.
   */
 double sum();

 /**
   Concatenate all the element (providing they are texts)
   and use the text given as argument as separator.
   */
 text join(const char *);

 int nomatch();

 //===== I/O

 /**
   Load a file into the dict. The argument is the filename.
   Each line of the file become an element. The end of line code
   (10, 13 or the twice) are stored also.
   */
 void load(const char *);

 /**
   Save the whole dict into the file, the name of which is given as
 argument. An end of line must be added to each element if necessary.
  */
 bool store(const char *);


 dict *to(int);
 dict *to(const char *);
 dict *to(text);

 dyn &at(int);
 dyn &at(text);
 dyn &at(const char *);
 dyn &at(dyn);
 
 void setAt(int, dyn *);
 void setAt(int, dyn);

 const char *toString(const char *s=", ");
 text toText(const char *s=", ");
 void boundError(const text &);

};


inline dict operator+(dict *a, const dict &b) { return a->operator+(b); }

std::ostream& operator<< (std::ostream& out, dict *);
std::ostream& operator<< (std::ostream& out, dict);

#endif

