/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.com.
*/


/**
   Sax - scriptol.com
   C++ wrapper to expat - libexpat.org
   Licence: see the expat archive.

   Build an automaton from callbacks that call it with a code
   according to each action the callback is intended to.
   Requires:
   - the sax lib for Visual C++ or BCC 5.5 or GCC.
   - the scriptol lib compatible with the same C++ compiler
     if you want to build a dom tree.
   How to use:
   - Create a class that inherits from sax
   - and override the automaton method.
   */


#include "expat.h"

enum { SAX_UNKNOW, SAX_START, SAX_END, SAX_DATA, SAX_COMMENT };

class sax {
 public:

 static int depth;
 static int depthold;

 static const char *elementName;    // set by element callback
 static const char *commentText;    // set by comment callback
 static const char **attributes;    // set by element callback
 static const char *dataText;       // set by text callback
 static const char *attributeName;  // set by the user
 static char *attributeValue;       // set at parsing
 static char *searchName;           // set by the user
 static char *foundValue;           // set at parsing
 static int len;
 static bool found;

 static void XMLCALL startElement(void *, const char *, const char **);

 static void XMLCALL endElement(void *, const char *);

 // Test if a real content, this tools handle any char outside elements!

 static bool valid(const char *, int);

 // Displaying the content

 static void XMLCALL content(void *, const char *, int);

 // Handling a comment

 static void XMLCALL comment(void *, const char *);

 // Parse the whole document and throw events
 // A method of class inherited of sax must be given as first parameter
 // belong with the name of a xml file the second parameter.
 // (you can assign directly the automaton pointer instead).

 static bool XMLCALL parse(const char *);
 static bool XMLCALL parse(void (*automaton) (int), const char *);

 // This is called by each callback with the corresponding code (see above)
 // Isr assigned a function by the parse method

 static void (* automaton) (int);

 // Display a document - a valid parameter for the parse method

 static void display(int);

 // Find an attribute from attributeName and set attributeValue if found
 // - set found to false before calling parse()
 // - set attributeName an attribute to find the value or
 // - set searchName an element name to find the data
 // - get the attribute value or element data from foundValue after parse().

 static void find(int);

};


extern dom *lxdom;
void _sax_build(int);
bool SaxImport(const char *);
