/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.com.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <fstream>
#include <iostream>

#ifndef UNIX
#define strcasecmp _stricmp
#endif

#include "text.hpp"

void text::reserve(int l) {
	data = (char *) new char[l];
}  

void text::set(const char *s) {
	len = strlen(s);
	data = (char *) new char[len + 1];
	memcpy(data, s, len + 1);
}

void text::set(text *s) { 
	set(s->data); 
}

void text::set(const text &t) { 
	set(t.data); 
}

void text::set(bool b) { 
	if(b) set("1"); 
	else set(""); 
}

void text::set(char c) {
	char s[2];
	s[0] = c; 
	s[1] = '\0'; 
	set(s);
}

void text::set(int n) {
  char buffer[64];
  sprintf((char *) buffer, "%d", n);
  set((const char *) buffer);
}

void text::set(natural n) {
  set((double) n);
}

void text::set(unsigned int n) {
  char buffer[64];
  sprintf((char *) buffer, "%u", n);
  set(buffer);
}


void text::set(double d)
{
  char buffer[128];
  sprintf((char *) buffer, "%g", d);
  set(buffer);
}

text::text() {
  len = 0;
  data = (char *) new char[1];
  data[0] = '\0';
};

text::text(const char *s) { set(s); }
text::text(const text &t) { set(t.data); }
text::text(text *x) { set(x->data); }
text::text(bool b) { set(b); }
text::text(char c) { set(c); }
text::text(int n) { set(n); }
text::text(natural n) { set(n); }
text::text(unsigned int n) { set(n); }
text::text(double d) { set(d); }
text::text(void *v) { set((char *) v); }
text::~text() {  delete data; }

text text::lower()
{
  text t;
  t.data = (char *) new char[len + 1];
  for(int i = 0; i < len; i++)
  {
    t.data[i] = (char) tolower(data[i]);
  }
  t.data[len] = '\0';
  t.len = len;
  return t;
}

text text::upper()
{
  text t;
  t.data = (char *) new char[len + 1];
  for(int i = 0; i < len; i++)
  {
    t.data[i] = (char) toupper(data[i]);
  }
  t.data[len] = '\0';
  t.len = len;
  return t;
}


char *text::toStr() {
  static char *x = (char *) new char[len + 1];
  memcpy(x, data, len + 1);
  return x;
}

text text::capitalize() {
  text b(data);
  if(len)  {
   b.data[0] = (char) toupper(b.data[0]);
  }
  return b;
}

// extract from stt to lst included

const char *text::substr(int stt, int lst)
{
  int l;
  char *n;

  if(empty()) goto bye;
  if(lst >= len) lst = len - 1;
  l = lst - stt + 1;
  if(l <= 0) goto bye;
  n = (char *) new char[l + 1];
  strncpy(n , data + stt, l);
  n[l] = '\0';
  return n;

bye:
  return "";
}

const char *text::substr(int s) { return substr(s, len - 1); }

void text::cat(const char *s) {
  int nl = len + strlen(s);

  char *x = (char *) new char[len + 1];
  memcpy(x, data, len + 1);

  data = (char *) realloc(data, nl + 1);
  memcpy(data, x, len + 1);
  strcat(data, s);
  len = nl;
}

void text::insert(int pos, const char *s) {
	text t;
	if(pos <= 0) {
		t.set(data);
		set(s);
		cat(t.data);
		return;
	}
	if(pos >= len) 	{
		cat(s);
		return;
	}
	t.set(substr(0, pos - 1));
	t.cat(s);
	t.cat(substr(pos));
	set(t);
}

bool text::in(const text &t) { return find(t.data, 0) != -1; }
bool text::in(const char *s) { return find(s, 0) != -1; }

int text::compareLex(const char *a, const char *b, int l) {
  if(strlen(a) < (unsigned) l) return -1;
  if(strlen(b) < (unsigned) l) return 1;
  for(int i = 0; i < l; i++)  {
    if(tolower(a[i]) != tolower(b[i]) )  return a[i] - b[i];
  }
  return 0;
}

int text::findLex(const char *sea, int stt) {
  int nl = strlen(sea);
  int l = len - nl + 1;
  for(int i = stt; i < l; i++)  {
    if(compareLex(data + i, sea, nl) == 0) return i;
  }
  return -1;
}

int text::find(const char *sea, int stt) {
  int nl = strlen(sea);
  int l = len - nl + 1;
  for(int i = stt; i < l; i++)  {
    if(strncmp(data + i, sea, nl) == 0) return i;
  }
  return -1;
}

int text::findLast(const char *sea, int stt) {
  int nl = strlen(sea);
  int l = len - nl + 1;
  for(int i = l; i >= 0; i--)  {
    if(strncmp(data + i, sea, nl) == 0) return i;
  }
  return -1;
}

/*
  - search for a substring: pos is the position
  - copy from start or end of already copied part until pos
  - copy the replacing string
  - move pos beyond the replaced substring
  - loop
*/

void text::replace(const char *fnd, const char *rep)
{
 int fl, rl, nl;
 int pos, soff, diff, tsize;
 char *n;

 fl = strlen(fnd);
 rl = strlen(rep);
 nl = len;
 pos = 0;       // offset in source of position to search

loop:
  pos = find(fnd, pos);
  if(pos != -1)
  {
   nl -= fl;
   nl += rl;
   pos += fl;
   goto loop;
  }
if(nl < 1) return;
n = (char *) new char[nl + 1];

soff = 0;      // offset in source of part to copy
tsize = 0;     // size of target

loop2:
 pos = find(fnd, soff);
 if(pos != -1)
 {
   diff = pos - soff;                      // size of old string to copy
   strncpy(n + tsize, data + soff, diff);   // copy what is kept from old string
   tsize += diff;          // add size of remaining part from source
   soff  += diff;          // add also to source offset
   strncpy(n + tsize, rep, rl); // copy replacing string
   tsize += rl;            // add size of replacING string
   soff += fl;             // add size of replacED string
   goto loop2;
 }
 strcpy(n + tsize, data + soff);
 //if(((tsize + strlen(data + soff)) != nl) printf("Error in replace %d %d\n", tsize, nl);
 n[nl] = '\0';
 set(n);
}

text text::slice(int stt, int lst, int incl)
{
  text x;

  if(len == 0) goto bye;
  if(lst == 0 && stt ==0) goto bye;

  if(!incl) if(lst) lst--;
  if(stt < 0) stt += len;
  if(stt < 0) stt = 0;
  if(stt >= len) goto bye;

  if(lst < 0) lst += len;
  if(lst < 0) lst = 0;
  if(lst >= len) lst = len - 1;

  lst -= stt;
  lst++;
  if(lst <= 0) goto bye;
  x.data = (char *) new char[lst + 1];
  strncpy(x.data, data + stt, lst);
  x.data[lst] = '\0';
  x.len = lst;

bye:
  return x;
}

void text::splice(int stt, int lst, const char *x, int incl)
 {
  int ls;
  char *s;

  if(len == 0) { set(x); return; }

  ls = strlen(x);

  lst += incl;
  lst--;

  if(lst < 0) lst = len + lst;
  if(lst < 0) return;
  if(lst >= len) lst = len - 1;

  if(stt < 0) stt = len + stt;
  if(stt < 0) stt = 0;
  if((lst + 1) < stt) return;
  if(stt >= len) { cat(x); return; }

  s = (char *) new char[stt + ls + len - lst + 1];
  strncpy(s, data, stt);
  strncpy(s + stt, x, ls);
  strcpy(s + stt + ls, data + lst + 1);
  set(s);
 }

void text::splice(int a, int b, const text &t, int i)
{
  splice(a,b,t.data, i);
}


text text::rTrim()
{
	int c;
	int i;
	int l = len;

	for (i = l - 1; i >= 0; i--)
	{
		c = (int) data[i];
		if (c <= 32) 
			l--;
		else 
			break;
	}
	return substr(0, l - 1);
}

text text::lTrim()
{
  char c;
  int i;

  for (i = 0; i < len; i++)
  {
    c = data[i];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') continue;
    else break;
  }
  return text(substr(i, len - 1));
}

text text::trim()
{
  char c;
  int l, i, j;

  for (i = 0; i < len; i++)
  {
    c = data[i];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') continue;
    else break;
  }
  l = len;
  for (j = len - 1; j >= i; j--)
  {
    c = data[j];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') l--;
    else  break;
  }
  return text(substr(i, l - 1));
}

void text::fill(const char *s, int n) { set((const text &) text(s).dup(n)); }

text text::dup(int n)
{
 int size = 0;
 text t;

 t.data = (char *) new char[len * n + 1];
 while(n > 0)
   {
     strncpy(t.data + size, data, len);
     size += len;
     n--;
   }

 t.len = size;
 t.data[size] = '\0';

 return t;
}

bool text::isNumber() { return isNumber(data); }
bool text::isNumber(const char *x)
 {
   int i = 0;
   int ctr = 0;
   char c;
   if(x == NULL) goto bye;
   do
   {
     c = x[i++];
     if(c == '\0') return true;
     if(c == 'e')
     {
       if(ctr) break;
       if(i > 1) { ctr++; continue; }
       break;
     }
     if(c < '0')   break;
     if(c > '9')   break;
   } while(1);
bye:
   return false;
 }


int text::compare(const char *s)
{

	double a, b;
	int l = strlen(s);

	if((len == l) && isNumber())
	{
		if(isNumber(s))
		{
			a = atof(data);
			b = atof(s);
			if(a == b) return 0;
			if(a < b) return -1;
			return 1;
		}
	}
	return strcasecmp(data, s);
 }

int text::identical(const char *s) 
{ 
	double a, b;
	int l = strlen(s);

	if((len == l) && isNumber())
	{
		if(isNumber(s))
		{
			a = atof(data);
			b = atof(s);
			if(a == b) return 0;
			if(a < b) return -1;
			return 1;
		}
	}
	int x = strcmp(data, s);
	if (x < 0) return -1;
	if (x > 0) return 1;
	return 0;
 }

char text::toChar()       { return data[0]; }
int text::toInt()         { return atoi(data); }
int text::toInteger()     { return atoi(data); }

double text::toNumber(const char *s)
{
	if(strlen(s) == 0) return (double) 0;
	return atof(s);
}

double text::toNumber()
 {
	if(length() == 0) return (double) 0;
	return atof(data);
 }
 
unsigned long text::toNatural() { 
  return (unsigned long) toNumber(); 
}

/**
   Save the text into a file.
*/
void text::toFile(const char *fname) {
  std::ofstream f;
  f.open(fname);
  f << data;
  f.close();
}
 


bool text::operator==(const text &x) { return identical(x.data) == 0; }
bool text::operator!=(const text &x) { return identical(x.data) != 0; }
bool text::operator<=(const text &x) { return identical(x.data) <= 0; }
bool text::operator>=(const text &x) { return identical(x.data) >= 0; }
bool text::operator<(const text &x)  { return identical(x.data) < 0; }
bool text::operator>(const text &x)  { return identical(x.data) > 0; }
bool text::operator==(const char *x) { return identical(x) == 0; }
bool text::operator!=(const char *x) { return identical(x) != 0; }
bool text::operator<=(const char *x) { return identical(x) <= 0; }
bool text::operator>=(const char *x) { return identical(x) >= 0; }
bool text::operator<(const char *x)  { return identical(x) < 0; }
bool text::operator>(const char *x)  { return identical(x) > 0; }
bool text::operator!()  { return length() == 0; }

bool text::operator==(bool x)
{
  bool b = true;
  if(length() == 0) b = false;
  if(isNumber() && (toNumber() == (double) 0)) b = false;
  return b == x;
}
bool text::operator!=(bool x) { return !operator==(x); }

bool text::operator==(int x)
{
  if(isNumber())
  {
    return toNumber() == (double) x;
  }
  return false;
}

bool text::operator!=(int x) { return !operator==(x); }

bool text::operator==(double x)
{
  if(isNumber())
  {
    return toNumber() == x;
  }
  return false;
}

bool text::operator!=(double x) { return !operator==(x); }


void text::operator=(const text &x) { set(x.data); }
void text::operator=(const char *x) { set(x);  }

text text::operator+(const text &x) { 
    text b = text(data); 
    b.cat(x.data); 
    return b; 
}

text text::operator+(const char *x) { 
    text b = text(data); 
    b.cat(x); 
    return b; 
}

void text::operator+=(const text &x) { 
    cat(x.data);
}

void text::operator+=(const char *x) { 
    cat(x); 
}

text &text::operator[](int n)
{
  if(n < len)
     op = new text(data[n]);
  else
     op = new text("");
  return (text &) *(op);
}

void text::setAt(int n, const text &t)
{
  if(t.len < 1) return;
  if(n < len)
  {
    data[n] = t.data[0];
    return;
  }
  cat(t.data);
}

text text::operator*(int n) { return dup(n); }
std::ostream& operator<< (std::ostream& out, text t)  { return out << t.data; }
//std::ostream& operator<< (std::ostream& out, text *t) { return out << t->data; }
